part of pdf;

class _Utils {
  static const List<int> def_reverse_bits = <int>[
    0,
    8,
    4,
    12,
    2,
    10,
    6,
    14,
    1,
    9,
    5,
    13,
    3,
    11,
    7,
    15
  ];
  static int bitReverse(int value) {
    return (def_reverse_bits[(value & 15)] << 12 |
            def_reverse_bits[((value >> 4) & 15)] << 8 |
            def_reverse_bits[((value >> 8) & 15)] << 4 |
            def_reverse_bits[(value >> 12)])
        .toSigned(16);
  }
}

class _CompressedStreamWriter {
  _CompressedStreamWriter(List<int> outputStream, bool bNoWrap,
      PdfCompressionLevel level, bool bCloseStream) {
    ArgumentError.checkNotNull(outputStream, 'outputStream');
    _treeLiteral = _CompressorHuffmanTree(
        this, def_huffman_literal_alphabet_length, 257, 15);
    _treeDistances = _CompressorHuffmanTree(
        this, def_huffman_distances_alphabet_length, 1, 15);
    _treeCodeLengths =
        _CompressorHuffmanTree(this, def_huffman_bitlen_tree_length, 4, 7);
    _arrDistancesBuffer = List<int>.filled(def_huffman_buffer_size, 0);
    _arrLiteralsBuffer = List<int>.filled(def_huffman_buffer_size, 0);
    _stream = outputStream;
    _level = level;
    _bNoWrap = bNoWrap;
    _bCloseStream = bCloseStream;
    _dataWindow = List<int>.filled(2 * wsize, 0);
    _hashHead = List<int>.filled(hash_size, 0);
    _hashPrevious = List<int>.filled(wsize, 0);
    _blockStart = _stringStart = 1;

    _goodLength = good_length[_getCompressionLevel(level)];
    _niceLength = nice_length[_getCompressionLevel(level)];
    _maximumChainLength = max_chain[_getCompressionLevel(level)];

    if (!bNoWrap) {
      _writeZLIBHeader();
    }
    initializeStaticLiterals();
  }

  /// Start template of the zlib header.
  static const int def_zlib_header_template = (8 + (7 << 4)) << 8;

  /// Memory usage level.
  static const int default_mem_level = 8;

  /// Size of the pending buffer.
  static const int def_pending_buffer_size = 1 << (default_mem_level + 8);

  /// Size of the buffer for the huffman encoding.
  static const int def_huffman_buffer_size = 1 << (default_mem_level + 6);

  /// Length of the literal alphabet(literal+lengths).
  static const int def_huffman_literal_alphabet_length = 286;

  /// Distances alphabet length.
  static const int def_huffman_distances_alphabet_length = 30;

  /// Length of the code-lengths tree.
  static const int def_huffman_bitlen_tree_length = 19;

  /// Code of the symbol, than means the end of the block.
  static const int def_huffman_endblock_symbol = 256;

  static const int too_far = 4096;

  /// Maximum window size.
  static const int wsize = 1 << 15;

  /// Internal compression engine constant
  static const int wmask = wsize - 1;

  /// Internal compression engine constant
  static const int hash_bits = default_mem_level + 7;

  /// Internal compression engine constant
  static const int hash_size = 1 << hash_bits;

  /// Internal compression engine constant
  static const int hash_mask = hash_size - 1;

  /// Internal compression engine constant
  static const int max_match = 258;

  /// Internal compression engine constant
  static const int min_match = 3;

  /// Internal compression engine constant
  static const int hash_shift = 5;

  /// Internal compression engine constant
  static const int min_lookahead = max_match + min_match + 1;

  /// Internal compression engine constant
  static const int max_dist = wsize - min_lookahead;

  /// Internal compression engine constant
  static const List<int> good_length = <int>[0, 4, 4, 4, 4, 8, 8, 8, 32, 32];

  /// Internal compression engine constant
  static const List<int> nice_length = <int>[
    0,
    8,
    16,
    32,
    16,
    32,
    128,
    128,
    258,
    258
  ];

  /// Internal compression engine constant
  static const List<int> max_chain = <int>[
    0,
    4,
    8,
    32,
    16,
    32,
    128,
    256,
    1024,
    4096
  ];

  static const List<int> def_huffman_dyntree_codelengths_order = <int>[
    16,
    17,
    18,
    0,
    8,
    7,
    9,
    6,
    10,
    5,
    11,
    4,
    12,
    3,
    13,
    2,
    14,
    1,
    15
  ];

  final List<int> _pendingBuffer = List<int>.filled(def_pending_buffer_size, 0);
  int _pendingBufferLength = 0;
  List<int> _arrDistancesBuffer;
  List<int> _arrLiteralsBuffer;
  List<int> _stream;
  PdfCompressionLevel _level;
  bool _bNoWrap = false;
  bool _bCloseStream = false;
  List<int> _dataWindow;
  List<int> _hashHead;
  List<int> _hashPrevious;
  int _blockStart = 0;
  int _stringStart = 0;
  int _lookAhead = 0;
  int _maximumChainLength = 0;
  int _niceLength = 0;
  int _goodLength = 0;
  List<int> _inputBuffer;
  int _inputOffset = 0;
  int _inputEnd = 0;
  bool _bStreamClosed = false;
  int _checksum = 1;
  int _pendingBufferBitsInCache = 0;
  int _pendingBufferBitsCache = 0;
  int _currentHash = 0;
  int _matchStart = 0;
  int _matchLength = 0;
  int _iBufferPosition = 0;
  bool _matchPreviousAvailable = false;
  _CompressorHuffmanTree _treeLiteral;
  _CompressorHuffmanTree _treeDistances;
  _CompressorHuffmanTree _treeCodeLengths;
  int _iExtraBits = 0;
  static List<int> _arrLiteralCodes;
  static List<int> _arrLiteralLengths;
  static List<int> _arrDistanceCodes;
  static List<int> _arrDistanceLengths;
  bool get _needsInput => _inputEnd == _inputOffset;
  bool get _pendingBufferIsFlushed => _pendingBufferLength == 0;
  bool get _huffmanIsFull => _iBufferPosition >= def_huffman_buffer_size;

  int _getCompressionLevel(PdfCompressionLevel level) {
    switch (level) {
      case PdfCompressionLevel.none:
        return 0;
      case PdfCompressionLevel.bestSpeed:
        return 1;
      case PdfCompressionLevel.belowNormal:
        return 3;
      case PdfCompressionLevel.normal:
        return 5;
      case PdfCompressionLevel.aboveNormal:
        return 7;
      case PdfCompressionLevel.best:
        return 9;
      default:
        return null;
    }
  }

  void initializeStaticLiterals() {
    if (_arrLiteralCodes == null) {
      _arrLiteralCodes =
          List<int>.filled(def_huffman_literal_alphabet_length, 0);
      _arrLiteralLengths =
          List<int>.filled(def_huffman_literal_alphabet_length, 0);
      int i = 0;
      while (i < 144) {
        _arrLiteralCodes[i] = _Utils.bitReverse((0x030 + i) << 8);
        _arrLiteralLengths[i++] = 8;
      }
      while (i < 256) {
        _arrLiteralCodes[i] = _Utils.bitReverse(((0x190 - 144) + i) << 7);
        _arrLiteralLengths[i++] = 9;
      }
      while (i < 280) {
        _arrLiteralCodes[i] = _Utils.bitReverse(((0x000 - 256) + i) << 9);
        _arrLiteralLengths[i++] = 7;
      }
      while (i < def_huffman_literal_alphabet_length) {
        _arrLiteralCodes[i] = _Utils.bitReverse(((0x0c0 - 280) + i) << 8);
        _arrLiteralLengths[i++] = 8;
      }
      _arrDistanceCodes =
          List<int>.filled(def_huffman_distances_alphabet_length, 0);
      _arrDistanceLengths =
          List<int>.filled(def_huffman_distances_alphabet_length, 0);

      for (i = 0; i < def_huffman_distances_alphabet_length; i++) {
        _arrDistanceCodes[i] = _Utils.bitReverse(i << 11);
        _arrDistanceLengths[i] = 5;
      }
    }
  }

  void _writeZLIBHeader() {
    // Initialize header.
    int iHeaderData = def_zlib_header_template;
    // Save compression level.
    iHeaderData |= ((_getCompressionLevel(_level) >> 2) & 3) << 6;
    // Align header.
    iHeaderData += 31 - (iHeaderData % 31);
    // Write header to stream.
    _pendingBufferWriteShortMSB(iHeaderData);
  }

  void _pendingBufferWriteShortMSB(int s) {
    _pendingBuffer[_pendingBufferLength++] = (s >> 8).toUnsigned(8);
    _pendingBuffer[_pendingBufferLength++] = s.toUnsigned(8);
  }

  void write(List<int> data, int offset, int length, bool bCloseAfterWrite) {
    ArgumentError.checkNotNull(data, 'data');
    final int end = offset + length;
    if (0 > offset || offset > end || end > data.length) {
      throw Exception('Offset or length is incorrect.');
    }
    _inputBuffer = data;
    _inputOffset = offset;
    _inputEnd = end;
    if (length == 0) {
      return;
    }
    if (_bStreamClosed) {
      throw Exception('Stream was closed.');
    }
    _checksum = _ChecksumCalculator._checksumUpdate(
        _checksum, _inputBuffer, offset, length);

    while (!_needsInput || !_pendingBufferIsFlushed) {
      _pendingBufferFlush();
      if (!_compressData(bCloseAfterWrite) && bCloseAfterWrite) {
        _pendingBufferFlush();
        _pendingBufferAlignToByte();

        if (!_bNoWrap) {
          _pendingBufferWriteShortMSB(_checksum >> 16);
          _pendingBufferWriteShortMSB(_checksum & 0xffff);
        }
        _pendingBufferFlush();
        _bStreamClosed = true;
        if (_bCloseStream) {
          _stream.clear();
        }
      }
    }
  }

  bool _compressData(bool finish) {
    bool success;
    do {
      _fillWindow();

      final bool canFlush = finish && _needsInput;

      success = _compressSlow(canFlush, finish);
    } while (_pendingBufferIsFlushed && success);

    return success;
  }

  bool _compressSlow(bool flush, bool finish) {
    if (_lookAhead < min_lookahead && !flush) {
      return false;
    }

    while (_lookAhead >= min_lookahead || flush) {
      if (_lookAhead == 0) {
        if (_matchPreviousAvailable) {
          _huffmanTallyLit(_dataWindow[_stringStart - 1] & 0xff);
        }

        _matchPreviousAvailable = false;

        _huffmanFlushBlock(
            _dataWindow, _blockStart, _stringStart - _blockStart, finish);

        _blockStart = _stringStart;

        return false;
      }

      if (_stringStart >= 2 * wsize - min_lookahead) {
        _slideWindow();
      }

      final int prevMatch = _matchStart;
      int prevLen = _matchLength;

      if (_lookAhead >= min_match) {
        final int hashHead = _insertString();

        if (hashHead != 0 &&
            _stringStart - hashHead <= max_dist &&
            _findLongestMatch(hashHead)) {
          /// Discard match if too small and too far away.
          if (_matchLength <= 5 &&
              (_matchLength == min_match) &&
              _stringStart - _matchStart > too_far) {
            _matchLength = min_match - 1;
          }
        }
      }

      /// Previous match was better.
      if (prevLen >= min_match && _matchLength <= prevLen) {
        _huffmanTallyDist(_stringStart - 1 - prevMatch, prevLen);
        prevLen -= 2;

        do {
          _stringStart++;
          _lookAhead--;

          if (_lookAhead >= min_match) {
            _insertString();
          }
        } while (--prevLen > 0);

        _stringStart++;
        _lookAhead--;
        _matchPreviousAvailable = false;
        _matchLength = min_match - 1;
      } else {
        if (_matchPreviousAvailable) {
          _huffmanTallyLit(_dataWindow[_stringStart - 1] & 0xff);
        }

        _matchPreviousAvailable = true;
        _stringStart++;
        _lookAhead--;
      }

      if (_huffmanIsFull) {
        int len = _stringStart - _blockStart;
        if (_matchPreviousAvailable) {
          len--;
        }

        final bool lastBlock =
            finish && _lookAhead == 0 && !_matchPreviousAvailable;
        _huffmanFlushBlock(_dataWindow, _blockStart, len, lastBlock);
        _blockStart += len;

        return !lastBlock;
      }
    }

    return true;
  }

  bool _findLongestMatch(int curMatch) {
    int chainLength = _maximumChainLength;
    int scan = _stringStart;
    int match;
    int bestEnd = _stringStart + _matchLength;
    int bestLen = max(_matchLength, min_match - 1);

    final int limit = max(_stringStart - max_dist, 0);

    final int strend = _stringStart + max_match - 1;
    int scanEnd1 = _dataWindow[bestEnd - 1];
    int scanEnd = _dataWindow[bestEnd];

    /// Do not waste too much time if we already have a good match.
    if (bestLen >= _goodLength) {
      chainLength >>= 2;
    }

    /// Do not look for matches beyond the end of the input. This is necessary to make deflate deterministic.

    if (_niceLength > _lookAhead) {
      _niceLength = _lookAhead;
    }

    do {
      if (_dataWindow[curMatch + bestLen] != scanEnd ||
          _dataWindow[curMatch + bestLen - 1] != scanEnd1 ||
          _dataWindow[curMatch] != _dataWindow[scan] ||
          _dataWindow[curMatch + 1] != _dataWindow[scan + 1]) {
        continue;
      }

      match = curMatch + 2;
      scan += 2;

      /// We check for insufficient _lookAhead only every 8th comparison and the 256th check will be made at _stringStart + 258.

      while (_dataWindow[++scan] == _dataWindow[++match] &&
          _dataWindow[++scan] == _dataWindow[++match] &&
          _dataWindow[++scan] == _dataWindow[++match] &&
          _dataWindow[++scan] == _dataWindow[++match] &&
          _dataWindow[++scan] == _dataWindow[++match] &&
          _dataWindow[++scan] == _dataWindow[++match] &&
          _dataWindow[++scan] == _dataWindow[++match] &&
          _dataWindow[++scan] == _dataWindow[++match] &&
          scan < strend) {}

      if (scan > bestEnd) {
        _matchStart = curMatch;
        bestEnd = scan;
        bestLen = scan - _stringStart;

        if (bestLen >= _niceLength) {
          break;
        }

        scanEnd1 = _dataWindow[bestEnd - 1];
        scanEnd = _dataWindow[bestEnd];
      }
      scan = _stringStart;
    } while ((curMatch = _hashPrevious[curMatch & wmask] & 0xffff) > limit &&
        --chainLength != 0);

    _matchLength = min(bestLen, _lookAhead);

    return _matchLength >= min_match;
  }

  bool _huffmanTallyDist(int dist, int len) {
    _arrDistancesBuffer[_iBufferPosition] = dist.toSigned(16);
    _arrLiteralsBuffer[_iBufferPosition++] = (len - 3).toUnsigned(8);

    final int lc = _huffmanLengthCode(len - 3);
    _treeLiteral._codeFrequences[lc]++;
    if (lc >= 265 && lc < 285) {
      _iExtraBits += (lc - 261) ~/ 4;
    }

    final int dc = _huffmanDistanceCode(dist - 1);
    _treeDistances._codeFrequences[dc]++;
    if (dc >= 4) {
      _iExtraBits += dc ~/ 2 - 1;
    }
    return _huffmanIsFull;
  }

  void _huffmanFlushBlock(
      List<int> stored, int storedOffset, int storedLength, bool lastBlock) {
    _treeLiteral._codeFrequences[def_huffman_endblock_symbol]++;

    // Build trees.
    _treeLiteral._buildTree();
    _treeDistances._buildTree();

    // Calculate bitlen frequency.
    _treeLiteral._calcBLFreq(_treeCodeLengths);
    _treeDistances._calcBLFreq(_treeCodeLengths);

    // Build bitlen tree.
    _treeCodeLengths._buildTree();

    int blTreeCodes = 4;
    for (int i = 18; i > blTreeCodes; i--) {
      if (_treeCodeLengths
              ._codeLengths[def_huffman_dyntree_codelengths_order[i]] >
          0) {
        blTreeCodes = i + 1;
      }
    }
    int optLen = 14 +
        blTreeCodes * 3 +
        _treeCodeLengths._getEncodedLength() +
        _treeLiteral._getEncodedLength() +
        _treeDistances._getEncodedLength() +
        _iExtraBits;

    int staticLen = _iExtraBits;
    for (int i = 0; i < def_huffman_literal_alphabet_length; i++) {
      staticLen += _treeLiteral._codeFrequences[i] * _arrLiteralLengths[i];
    }
    for (int i = 0; i < def_huffman_distances_alphabet_length; i++) {
      staticLen += _treeDistances._codeFrequences[i] * _arrDistanceLengths[i];
    }
    if (optLen >= staticLen) {
      // Force static trees.
      optLen = staticLen;
    }

    if (storedOffset >= 0 && storedLength + 4 < optLen >> 3) {
      _huffmanFlushStoredBlock(stored, storedOffset, storedLength, lastBlock);
    } else if (optLen == staticLen) {
      // Encode with static tree.
      _pendingBufferWriteBits((1 << 1) + (lastBlock ? 1 : 0), 3);
      _treeLiteral._setStaticCodes(_arrLiteralCodes, _arrLiteralLengths);
      _treeDistances._setStaticCodes(_arrDistanceCodes, _arrDistanceLengths);
      _huffmanCompressBlock();
      _huffmanReset();
    } else {
      // Encode with dynamic tree.
      _pendingBufferWriteBits((2 << 1) + (lastBlock ? 1 : 0), 3);
      _huffmanSendAllTrees(blTreeCodes);
      _huffmanCompressBlock();
      _huffmanReset();
    }
  }

  void _huffmanSendAllTrees(int blTreeCodes) {
    _treeCodeLengths._buildCodes();
    _treeLiteral._buildCodes();
    _treeDistances._buildCodes();
    _pendingBufferWriteBits(_treeLiteral._codeCount - 257, 5);
    _pendingBufferWriteBits(_treeDistances._codeCount - 1, 5);
    _pendingBufferWriteBits(blTreeCodes - 4, 4);

    for (int rank = 0; rank < blTreeCodes; rank++) {
      _pendingBufferWriteBits(
          _treeCodeLengths
              ._codeLengths[def_huffman_dyntree_codelengths_order[rank]],
          3);
    }

    _treeLiteral._writeTree(_treeCodeLengths);
    _treeDistances._writeTree(_treeCodeLengths);
  }

  int _insertString() {
    int match;
    final int hash = ((_currentHash << hash_shift) ^
            _dataWindow[_stringStart + (min_match - 1)]) &
        hash_mask;

    _hashPrevious[_stringStart & wmask] = match = _hashHead[hash];
    _hashHead[hash] = _stringStart.toSigned(16);
    _currentHash = hash;
    return match & 0xffff;
  }

  void _huffmanCompressBlock() {
    for (int i = 0; i < _iBufferPosition; i++) {
      final int litlen = _arrLiteralsBuffer[i] & 255;
      int dist = _arrDistancesBuffer[i];

      if (dist-- != 0) {
        final int lc = _huffmanLengthCode(litlen);
        _treeLiteral._writeCodeToStream(lc);

        int bits = (lc - 261) ~/ 4;
        if (bits > 0 && bits <= 5) {
          _pendingBufferWriteBits(litlen & ((1 << bits) - 1), bits);
        }

        final int dc = _huffmanDistanceCode(dist);
        _treeDistances._writeCodeToStream(dc);

        bits = dc ~/ 2 - 1;
        if (bits > 0) {
          _pendingBufferWriteBits(dist & ((1 << bits) - 1), bits);
        }
      } else {
        _treeLiteral._writeCodeToStream(litlen);
      }
    }

    _treeLiteral._writeCodeToStream(def_huffman_endblock_symbol);
  }

  int _huffmanDistanceCode(int distance) {
    int code = 0;

    while (distance >= 4) {
      code += 2;
      distance >>= 1;
    }
    return code + distance;
  }

  int _huffmanLengthCode(int len) {
    if (len == 255) {
      return 285;
    }

    int code = 257;

    while (len >= 8) {
      code += 4;
      len >>= 1;
    }

    return code + len;
  }

  void _huffmanFlushStoredBlock(
      List<int> stored, int storedOffset, int storedLength, bool lastBlock) {
    _pendingBufferWriteBits((0 << 1) + (lastBlock ? 1 : 0), 3);
    _pendingBufferAlignToByte();
    _pendingBufferWriteShort(storedLength);
    _pendingBufferWriteShort(~storedLength);
    _pendingBufferWriteByteBlock(stored, storedOffset, storedLength);
    _huffmanReset();
  }

  void _huffmanReset() {
    _iBufferPosition = 0;
    _iExtraBits = 0;
    _treeLiteral._reset();
    _treeDistances._reset();
    _treeCodeLengths._reset();
  }

  void _pendingBufferWriteShort(int s) {
    _pendingBuffer[_pendingBufferLength++] = s.toUnsigned(8);
    _pendingBuffer[_pendingBufferLength++] = (s >> 8).toUnsigned(8);
  }

  void _pendingBufferWriteByteBlock(List<int> data, int offset, int length) {
    List.copyRange(
        _pendingBuffer, _pendingBufferLength, data, offset, offset + length);
    _pendingBufferLength += length;
  }

  void _pendingBufferAlignToByte() {
    if (_pendingBufferBitsInCache > 0) {
      _pendingBuffer[_pendingBufferLength++] =
          _pendingBufferBitsCache.toUnsigned(8);

      if (_pendingBufferBitsInCache > 8) {
        _pendingBuffer[_pendingBufferLength++] =
            (_pendingBufferBitsCache >> 8).toUnsigned(8);
      }
    }

    _pendingBufferBitsCache = 0;
    _pendingBufferBitsInCache = 0;
  }

  void _pendingBufferWriteBits(int b, int count) {
    _pendingBufferBitsCache |= (b << _pendingBufferBitsInCache).toUnsigned(32);
    _pendingBufferBitsInCache += count;

    _pendingBufferFlushBits();
  }

  bool _huffmanTallyLit(int literal) {
    _arrDistancesBuffer[_iBufferPosition] = 0;
    _arrLiteralsBuffer[_iBufferPosition++] = literal.toUnsigned(8);
    _treeLiteral._codeFrequences[literal]++;
    return _huffmanIsFull;
  }

  void _fillWindow() {
    if (_stringStart >= wsize + max_dist) {
      _slideWindow();
    }

    while (_lookAhead < min_lookahead && _inputOffset < _inputEnd) {
      int more = 2 * wsize - _lookAhead - _stringStart;
      if (more > _inputEnd - _inputOffset) {
        more = _inputEnd - _inputOffset;
      }
      List.copyRange(_dataWindow, _stringStart + _lookAhead, _inputBuffer,
          _inputOffset, _inputOffset + more);

      _inputOffset += more;
      _lookAhead += more;
    }

    if (_lookAhead >= min_match) {
      _updateHash();
    }
  }

  void _slideWindow() {
    List.copyRange(_dataWindow, 0, _dataWindow, wsize, wsize + wsize);
    _matchStart -= wsize;
    _stringStart -= wsize;
    _blockStart -= wsize;

    for (int i = 0; i < hash_size; ++i) {
      final int m = _hashHead[i] & 0xffff;
      _hashHead[i] = ((m >= wsize) ? (m - wsize) : 0).toSigned(16);
    }

    for (int i = 0; i < wsize; i++) {
      final int m = _hashPrevious[i] & 0xffff;
      _hashPrevious[i] = ((m >= wsize) ? (m - wsize) : 0).toSigned(16);
    }
  }

  void _updateHash() {
    _currentHash = (_dataWindow[_stringStart] << hash_shift) ^
        _dataWindow[_stringStart + 1];
  }

  void _pendingBufferFlush() {
    _pendingBufferFlushBits();
    if (_pendingBufferLength > 0) {
      final List<int> array = List<int>.filled(_pendingBufferLength, 0);
      array.setAll(0, _pendingBuffer.sublist(0, _pendingBufferLength));
      _stream.addAll(array);
    }
    _pendingBufferLength = 0;
  }

  /// Flushes fully recorded bytes to buffer array.
  int _pendingBufferFlushBits() {
    int result = 0;
    while (_pendingBufferBitsInCache >= 8 &&
        _pendingBufferLength < def_pending_buffer_size) {
      _pendingBuffer[_pendingBufferLength++] =
          _pendingBufferBitsCache.toUnsigned(8);
      _pendingBufferBitsCache >>= 8;
      _pendingBufferBitsInCache -= 8;
      result++;
    }
    return result;
  }

  // ignore: unused_element
  void _close() {
    if (_bStreamClosed) {
      return;
    }
    do {
      _pendingBufferFlush();
      if (!_compressData(true)) {
        _pendingBufferFlush();
        _pendingBufferAlignToByte();

        if (!_bNoWrap) {
          _pendingBufferWriteShortMSB(_checksum >> 16);
          _pendingBufferWriteShortMSB(_checksum & 0xffff);
        }

        _pendingBufferFlush();
      }
    } while (!_needsInput || !_pendingBufferIsFlushed);

    _bStreamClosed = true;

    if (_bCloseStream) {
      _stream.clear();
    }
  }
}

class _ChecksumCalculator {
  static const int checkSumBitOffset = 16;
  static const int checksumBase = 65521;
  static const int checksumIterationCount = 3800;

  static int _checksumUpdate(
      int checksum, List<int> buffer, int offset, int length) {
    int checksumUint = checksum.toUnsigned(32);
    int s1 = checksumUint & 65535;
    int s2 = checksumUint >> checkSumBitOffset;
    while (length > 0) {
      int steps = min(length, checksumIterationCount);
      length -= steps;
      while (--steps >= 0) {
        s1 = s1 + (buffer[offset++] & 255).toUnsigned(32);
        s2 += s1;
      }
      s1 %= checksumBase;
      s2 %= checksumBase;
    }
    checksumUint = (s2 << checkSumBitOffset) | s1;
    return checksumUint;
  }
}
