part of pdf;

class _PdfMainObjectCollection {
  //Constructor
  _PdfMainObjectCollection() {
    _initialize();
  }

  //Fields
  int _index;
  Map<int, _ObjectInfo> _mainObjectCollection;
  List<_ObjectInfo> _objectCollection;
  Map<_IPdfPrimitive, int> _primitiveObjectCollection;
  int _maximumReferenceObjectNumber;

  //Properties
  int get _count {
    return _objectCollection.length;
  }

  /// Get the PdfMainObjectCollection items.
  _ObjectInfo operator [](int key) => _returnValue(key);

  //Implementation
  void _initialize() {
    _index = 0;
    _mainObjectCollection = <int, _ObjectInfo>{};
    _objectCollection = <_ObjectInfo>[];
    _primitiveObjectCollection = <_IPdfPrimitive, int>{};
  }

  /// Check key and return the value.
  _ObjectInfo _returnValue(int index) {
    if (index < 0 || index > _objectCollection.length) {
      throw ArgumentError.value(index, 'index', 'index out of range');
    }
    return _objectCollection[index];
  }

  /// Adds the specified element.
  void _add(dynamic element, [_PdfReference reference]) {
    ArgumentError.checkNotNull(element, 'element');
    if (element is _IPdfWrapper) {
      element = element._element;
    }
    if (reference == null) {
      final _ObjectInfo info = _ObjectInfo(element);
      _objectCollection.add(info);
      if (!_primitiveObjectCollection.containsKey(element)) {
        _primitiveObjectCollection[element] = _objectCollection.length - 1;
      }
      element.position = _objectCollection.length - 1;
      _index = _objectCollection.length - 1;
      element.status = _ObjectStatus.registered;
    } else {
      final _ObjectInfo info = _ObjectInfo(element, reference);
      if (_maximumReferenceObjectNumber < reference._objNum) {
        _maximumReferenceObjectNumber = reference._objNum;
      }
      _objectCollection.add(info);
      if (!_primitiveObjectCollection.containsKey(element)) {
        _primitiveObjectCollection[element] = _objectCollection.length - 1;
      }
      _mainObjectCollection[reference._objNum] = info;
      element.position = _objectCollection.length - 1;
      reference.position = _objectCollection.length - 1;
    }
  }

  _PdfReference _getReference(_IPdfPrimitive object) {
    _index = _lookFor(object);
    final _ObjectInfo objectInfo = _objectCollection[_index];
    return objectInfo._reference;
  }

  int _lookFor(_IPdfPrimitive obj) {
    int index = -1;
    if (obj.position != -1) {
      return obj.position;
    }
    if (_primitiveObjectCollection.containsKey(obj) &&
        _count == _primitiveObjectCollection.length) {
      index = _primitiveObjectCollection[obj];
    } else {
      for (int i = _count - 1; i >= 0; i--) {
        final _ObjectInfo objectInfo = _objectCollection[i];
        if (objectInfo._object == obj) {
          index = i;
          break;
        }
      }
    }
    return index;
  }

  bool _trySetReference(_IPdfPrimitive object, _PdfReference reference) {
    ArgumentError.checkNotNull(object, 'object');
    ArgumentError.checkNotNull(reference, 'reference');
    bool result = true;
    _index = _lookFor(object);
    if (_index < 0 || _index >= _objectCollection.length) {
      result = false;
    } else {
      final _ObjectInfo objectInfo = _objectCollection[_index];
      if (objectInfo._reference != null) {
        result = false;
      } else {
        objectInfo._setReference(reference);
      }
    }
    return result;
  }

  void _dispose() {
    if (_mainObjectCollection != null) {
      _mainObjectCollection.clear();
      _mainObjectCollection = null;
    }
    if (_objectCollection != null) {
      _objectCollection.clear();
      _objectCollection = null;
    }
    if (_primitiveObjectCollection != null &&
        _primitiveObjectCollection.isNotEmpty) {
      final List<_IPdfPrimitive> primitives =
          _primitiveObjectCollection.keys.toList();
      for (int i = 0; i < primitives.length; i++) {
        primitives[i].dispose();
      }
      _primitiveObjectCollection.clear();
      _primitiveObjectCollection = null;
    }
  }
}
