part of pdf;

class _PdfStream extends _PdfDictionary {
  _PdfStream([_PdfDictionary dictionary, List<int> data]) {
    if (dictionary == null && data == null) {
      _dataStream = <int>[];
      _compress = true;
    } else {
      ArgumentError.checkNotNull(data, 'data');
      ArgumentError.checkNotNull(dictionary, 'dictionary');
      _compress = false;
      _dataStream = <int>[];
      _dataStream.addAll(data);
      this[_DictionaryProperties.length] = _PdfNumber(_dataStream.length);
    }
  }

  //Constants
  static const String prefix = 'stream';
  static const String suffix = 'endstream';

  //Fields
  List<int> _dataStream;
  bool _compress;
  bool _isChanged;

  //Properties
  bool get compress {
    return _compress;
  }

  set compress(bool value) {
    _compress = value;
    _modify();
  }

  //Implementations
  void _modify() {
    _isChanged = true;
  }

  List<int> _compressContent(_IPdfWriter writer) {
    final List<int> data = _dataStream;
    if (compress &&
        writer._document.compressionLevel != PdfCompressionLevel.none) {
      final List<int> outputStream = <int>[];
      final _CompressedStreamWriter compressedWriter = _CompressedStreamWriter(
          outputStream, false, writer._document.compressionLevel, false);
      compressedWriter.write(data, 0, data.length, false);
      compressedWriter._close();
      _addFilter(_DictionaryProperties.flateDecode);
      return outputStream;
    } else {
      return data;
    }
  }

  void _addFilter(String filterName) {
    _IPdfPrimitive filter = _items[_DictionaryProperties.filter];
    final _PdfReferenceHolder obj = filter as _PdfReferenceHolder;
    if (obj != null) {
      filter = obj.object;
    }
    _PdfArray array = filter as _PdfArray;
    _PdfName name = filter as _PdfName;
    if (name != null) {
      array = _PdfArray();
      array._insert(0, name);
      this[_DictionaryProperties.filter] = array;
    }
    name = _PdfName(filterName);
    if (array == null) {
      this[_DictionaryProperties.filter] = name;
    } else {
      array._insert(0, name);
    }
  }

  void _write(dynamic pdfObject) {
    ArgumentError.checkNotNull(pdfObject);
    if (pdfObject.isEmpty) {
      throw ArgumentError.value(pdfObject, 'value cannot be empty');
    }
    if (pdfObject is String) {
      _write(utf8.encode(pdfObject));
    } else if (pdfObject is List<int>) {
      _dataStream.addAll(pdfObject);
      _modify();
    } else {
      throw ArgumentError.value(
          pdfObject, 'The method or operation is not implemented');
    }
  }

  void _clearStream() {
    _dataStream.clear();
    if (containsKey(_DictionaryProperties.filter)) {
      remove(_DictionaryProperties.filter);
    }
    compress = true;
    _modify();
  }

  //_IPdfChangables members
  @override
  void freezeChanges(Object freezer) {}

  //_IPdfPrimitive members
  @override
  void save(_IPdfWriter writer) {
    final _SavePdfPrimitiveArgs beginSaveArguments =
        _SavePdfPrimitiveArgs(writer);
    _onBeginSave(beginSaveArguments);
    final List<int> data = _compressContent(writer);
    this[_DictionaryProperties.length] = _PdfNumber(data.length);
    super._saveDictionary(writer, false);
    writer._write(prefix);
    writer._write(_Operators.newLine);
    if (data.isNotEmpty) {
      writer._write(data);
      writer._write(_Operators.newLine);
    }
    writer._write(suffix);
    writer._write(_Operators.newLine);
    final _SavePdfPrimitiveArgs endSaveArguments =
        _SavePdfPrimitiveArgs(writer);
    _onEndSave(endSaveArguments);
    if (_compress) {
      remove(_DictionaryProperties.filter);
    }
  }

  @override
  void dispose() {
    if (_dataStream != null && _dataStream.isNotEmpty) {
      _dataStream.clear();
      _dataStream = null;
    }
  }
}
